/**         @file  assert.c
 *         @brief  Compiled functions used by debugging macros to write on stderr
 *          @date  11/05/1997
 *        @author  François Cerbelle (Fanfan), francois@cerbelle.net
 *     @copyright  Copyright (c) 1997-2024, François Cerbelle
 *
 *  Originally inspired by "L'art du code", Steve Maguire, Microsoft Press
 *
 *   @internal
 *       Compiler  gcc
 *  Last modified  2024-07-27 20:44
 *   Organization  Cerbelle.net
 *        Company  Home
 *
 *   This source code is released for free distribution under the terms of the
 *   GNU General Public License as published by the Free Software Foundation.
 *
 */

#include "assert.h"
#include <stdarg.h>                             /* va_list, va_start, va_arg, va_end */
#include <string.h>                             /* strcpy */
#include <stdio.h>                              /* fprintf */
#include <time.h>                               /* time, localtime */

/** Build a timestamp string with filename and compilation date
 *
 * Returned value is a static string which can be modified later. It should not
 * be passed to free and should not be modified.
 *
 * NULL or empty string parameter values will abort the program.
 *
 *  @param[in] p_File Source file
 *  @param[in] p_Line Source line in the source file
 *  @param[in] p_CompilDate Compilation date
 *  @param[in] p_CompilTime Compilation time
 *  @param[in] p_Function Function name in the source file
 *
 *  @return Pointer to a static string
 */
static const char*_timestamp(const char* p_File,
                             const unsigned int p_Line,
                             const char* p_CompilDate,
                             const char* p_CompilTime,
                             const char* p_Function)
{
    /* Get local time and format it */
    char l_Time[24];
    static char l_tmp[110];
    time_t l_CurrentTime = time(NULL);

    /* Parameter validity check against invalid parameters such as NULL or
     * empty string values. */
    if ((NULL==p_File)|| (0==p_File[0])||
        (0==p_Line)||
        (NULL==p_CompilDate)|| (0==p_CompilDate[0])||
        (NULL==p_CompilTime)|| (0==p_CompilTime[0])||
        (NULL==p_Function)|| (0==p_Function[0])) {
        fprintf(stderr,"%s:%d Unexpected and invalid parameters\n",__FILE__, __LINE__);
        abort();
    }

    strftime(l_Time, sizeof(l_Time), "%Y-%m-%d %H:%M:%S", localtime(&l_CurrentTime));

    /* Timestamp build with the filename, file line, function name */
    /** @todo Replace with a portable snprintf function */
    snprintf(l_tmp,sizeof(l_tmp),
             "%19s [%20s:%-4ud] (%11s @ %8s) %30s()",
             l_Time,p_File,p_Line,p_CompilDate,p_CompilTime,p_Function
            );
    return &l_tmp[0];
}

/* Documentation in header file */
void _trace(const char* p_File,
            const unsigned int p_Line,
            const char* p_CompilDate,
            const char* p_CompilTime, const char* p_Function)
{
    /* Parameter validity enforced by _timestamp */
    fprintf (stderr,"%s\n",
             _timestamp( p_File, p_Line, p_CompilDate, p_CompilTime, p_Function)
            );
}

/* Documentation in header file */
void _trace_msg(const char* p_File,
                const unsigned int p_Line,
                const char* p_CompilDate,
                const char* p_CompilTime,
                const char* p_Function, const char* p_Message)
{
    /* Parameter validity enforced by _timestamp */
    fprintf (stderr,"%s : %s\n",
             _timestamp( p_File, p_Line, p_CompilDate, p_CompilTime, p_Function),
             p_Message
            );
}

/* Documentation in header file */
void _trace_dynmsg(const char* p_File,
                   const unsigned int p_Line,
                   const char* p_CompilDate,
                   const char* p_CompilTime,
                   const char* p_Function, const char* p_Format, ...
                  )
{
    /* Formatted message string */
    char l_tmp[100];
    /* Formatted message length */
    unsigned int l_length;

    /* Limit variadic processing scope in a block */
    {
        /* Build the formatted message */
        va_list l_ap;
        va_start(l_ap, p_Format);
        /** @todo Replace with a portable snprintf function */
        l_length = vsnprintf(l_tmp,sizeof(l_tmp),p_Format,l_ap);
        va_end(l_ap);
    }

    if (l_length >= sizeof(l_tmp)-1) {
        /* Indicate that message was truncated */
        strcpy(&(l_tmp[sizeof(l_tmp)-6]),"[...]");
    }

    /* Parameter validity enforced by _timestamp */
    fprintf (stderr,"%s : %s\n",
             _timestamp( p_File, p_Line, p_CompilDate, p_CompilTime, p_Function),
             l_tmp
            );
}

